from zeroconf import Zeroconf, ServiceBrowser, ServiceStateChange
from typing import List, Dict

class BoseDiscovery:
  """ Discover Bose devices on the local network using Zeroconf. """
  def __init__(self, zeroconf: Zeroconf = None):
    if zeroconf is None:
      zeroconf = Zeroconf()
    self.zeroconf = zeroconf
    self.devices: List[Dict[str, str]] = []

  def _on_service_state_change(self, zeroconf, service_type, name, state_change):
    if state_change == ServiceStateChange.Added:
      self._resolve_service(name)

  def _resolve_service(self, name: str):
    info = self.zeroconf.get_service_info("_bose-passport._tcp.local.", name)
    if info:
      guid = info.properties.get(b"GUID")
      if guid:
        guid = guid.decode("utf-8")
      addresses = [addr for addr in info.parsed_addresses()]
      if addresses:
        self.devices.append({"GUID": guid, "IP": addresses[0]})

  def discover_devices(self, timeout: int = 5) -> List[Dict[str, str]]:
    """
    Discover devices in the `_bose-passport._tcp.local` service and return their GUID and IP address.

    :param timeout: Time in seconds to wait for discovery.
    :return: List of dictionaries containing GUID and IP for each discovered device.
    """
    self.devices = []

    listener = ServiceBrowser(self.zeroconf, "_bose-passport._tcp.local.", handlers=[self._on_service_state_change])

    try:
        import time
        time.sleep(timeout)
    finally:
        self.zeroconf.close()

    return self.devices

# EXAMPLE USAGE

if __name__ == "__main__":
  discovery = BoseDiscovery()
  devices = discovery.discover_devices()
  for device in devices:
      print(f"GUID: {device['GUID']}, IP: {device['IP']}")


"""
Bose Speaker control using its Websocket (like the BOSE app)

In order to control the device locally, you need to obtain the control token and device ID.
The control token needs to be aquired from the online BOSE API. The script "BoseAuth.py" can be used to obtain the control token using the email and password of the BOSE account.
The device ID can be obtained by discovering the device on the local network using the script "BoseDiscovery.py".

The token is only valid for a certain amount of time AND does not renew automatically.
So you may need to refetch the token from time to time.
"""

import json
import asyncio
import logging
from ssl import SSLContext, CERT_NONE
import websockets
from threading import Event
from .BoseResponse import AudioVolume, ContentNowPlaying, SystemInfo, SystemPowerControl, Sources, Audio, Accessories, Battery
import sys

# These are the default resources that are subscribed to when connecting to the speaker by the BOSE app
DEFAULT_SUBSCRIBE_RESOURCES = [
  "/bluetooth/sink/list",
  "/system/power/control",
  "/audio/avSync",
  "/bluetooth/source/list",
  "/audio/bass",
  "/device/assumed/TVs",
  "/network/wifi/siteScan",
  "/system/update/start",
  "/system/update/status",
  "/bluetooth/sink/macAddr",
  "/content/nowPlaying/rating",
  "/bluetooth/source/stopScan",
  "/system/setup",
  "/homekit/info",
  "/bluetooth/source/pairStatus",
  "/device/configuredDevices",
  "/bluetooth/source/status",
  "/cast/teardown",
  "/bluetooth/sink/status",
  "/cast/setup",
  "/cec",
  "/cloudSync",
  "/system/challenge",
  "/bluetooth/sink/remove",
  "/bluetooth/source/connect",
  "/remote/integration/brandList",
  "/subscription",
  "/network/status",
  "/bluetooth/source/scanResult",
  "/content/playbackRequest",
  "/audio/eqSelect",
  "/audio/height",
  "/content/transportControl",
  "/grouping/activeGroups",
  "/audio/mode",
  "/bluetooth/source/pair",
  "/bluetooth/source/capability",
  "/bluetooth/source/disconnect",
  "/audio/subwooferGain",
  "/voice/setup/start",
  "/audio/center",
  "/network/wifi/status",
  "/content/nowPlaying/repeat",
  "/system/sources",
  "/content/nowPlaying",
  "/system/power/macro",
  "/bluetooth/sink/pairable",
  "/network/wifi/profile",
  "/cast/settings",
  "/audio/zone",
  "/content/nowPlaying/shuffle",
  "/bluetooth/source/capabilitySettings",
  "/remote/integration",
  "/audio/surround",
  "/accessories",
  "/audio/treble",
  "/adaptiq",
  "/accessories/playTones",
  "/system/power/timeouts",
  "/audio/dualMonoSelect",
  "/system/info",
  "/system/sources/status",
  "/audio/rebroadcastLatency/mode",
  "/audio/format",
  "/bluetooth/source/connectionStatus",
  "/system/power/mode/opticalAutoWake",
  "/content/nowPlaying/favorite",
  "/system/productSettings",
  "/bluetooth/sink/connectionStatus",
  "/bluetooth/source/remove",
  "/audio/autoVolume",
  "/system/capabilities",
  "/audio/volume/increment",
  "/bluetooth/sink/connect",
  "/bluetooth/source/volume",
  "/bluetooth/sink/disconnect",
  "/system/reset",
  "/audio/volume/decrement",
  "/audio/volume",
  "/remote/integration/directEntry",
  "/device/configure",
  "/device/setup",
  "/bluetooth/source/scan",
  "/voice/settings",
  "/system/activated"
]


class BoseSpeaker:
    def __init__(self, control_token: str, host: str, device_id = None, version = 1):
        self._control_token = control_token
        self._device_id = device_id
        self._host = host
        self._version = version
        self._websocket = None
        self._ssl_context = SSLContext()
        self._ssl_context.verify_mode = CERT_NONE
        self._subprotocol = "eco2"
        self._req_id = 1
        self._url = f"wss://{self._host}:8082/?product=Madrid-iOS:31019F02-F01F-4E73-B495-B96D33AD3664"
        self._responses = []
        self._stop_event = Event()
        self._receiver_task = None
        self._receivers = {}
        self._subscribed_resources = []
        self._message_queue = asyncio.Queue()

    async def connect(self):
        """Connect to the WebSocket and start the receiver task."""
        self._websocket = await websockets.connect(
            self._url,
            subprotocols=[self._subprotocol],
            ssl=self._ssl_context
        )
        logging.info("WebSocket connection established.")

        self._stop_event.clear()
        self._receiver_task = asyncio.create_task(self._receiver_loop())
        if len(self._subscribed_resources) > 0:
            logging.debug("Subscribing to resources from previous session.")
            await self.subscribe(self._subscribed_resources)

    async def disconnect(self):
        """Stop the receiver task and close the WebSocket."""
        self._stop_event.set()
        if self._receiver_task:
            await self._receiver_task
        if self._websocket:
            await self._websocket.close()
        logging.info("WebSocket connection closed.")

    def attach_receiver(self, callback) -> int:
        """Attach to receiver."""
        id = max(self._receivers.keys(), default=0) + 1
        self._receivers[id] = callback
        return id

    def detach_receiver(self, id):
        """Detach from receiver."""
        self.receivers.pop(id, None)

    async def _request(self, resource, method, body={}, withHeaders=False, waitForResponse=True, version = None):
        """Send a request and wait for the matching response."""
        token = self._control_token
        req_id = self._req_id
        self._req_id += 1

        if version is None:
            version = self._version

        message = {
            "body": body,
            "header": {
                "token": token,
                "version": version,
                "reqID": req_id,
                "resource": resource,
                "device": self._device_id,
                "msgtype": "REQUEST",
                "method": method
            }
        }

        if self._device_id is None:
            await self._message_queue.put(message)
            logging.debug(f"Waiting for deviceID. Queued message: {json.dumps(message, indent=4)}")
        else:
            if self._websocket is None or self._websocket.close_code is not None:
                logging.warning("WebSocket connection is closed. Reconnecting before sending message.")
                await self.connect()
            await self._websocket.send(json.dumps(message))
            logging.debug(f"Sent message: {json.dumps(message, indent=4)}")

        # Wait for response with matching reqID
        if not waitForResponse:
            return

        # TODO: Refactor from polling to event-driven
        while True:
            for response in self._responses:
                if response["header"]["reqID"] == req_id:
                    self._responses.remove(response)
                    if "status" in response["header"] and response["header"]["status"] != 200:
                        raise Exception(f"Request failed with status {response['header']['status']}")
                    if not withHeaders:
                      return response["body"]
                    return response
            await asyncio.sleep(0.1)

    async def _receiver_loop(self):
        """Async function to receive and process messages."""
        try:
            while not self._stop_event.is_set():
                message = await self._websocket.recv()
                logging.debug(f"Received message: {message}")
                parsed_message = json.loads(message)

                if "header" in parsed_message and "device" in parsed_message["header"] and self._device_id is None:
                    self._device_id = parsed_message["header"]["device"]
                    while not self._message_queue.empty():
                        message = await self._message_queue.get()
                        message["header"]["device"] = self._device_id
                        await self._websocket.send(json.dumps(message))
                        logging.debug(f"Sent queued message: {json.dumps(message, indent=4)}")

                # Check if the message is a response to a request
                if "header" in parsed_message and "reqID" in parsed_message["header"]:
                    self._responses.append(parsed_message)
                else:
                    # Notify all receivers about the unsolicited message
                    for receiver in self._receivers.values():
                        receiver(parsed_message)

        except websockets.ConnectionClosed:
            logging.warning("WebSocket connection lost. Attempting to reconnect...")
            await self.connect()  # Reconnect if the connection is lost
            return
        except Exception as e:
            if not self._stop_event.is_set():
                logging.error(f"Error in receiver loop: {e}")

    async def get_capabilities(self):
        """Get the capabilities of the device."""
        return await self._request("/system/capabilities", "GET")

    async def get_system_info(self) -> SystemInfo:
        """Get system info."""
        return SystemInfo(await self._request("/system/info", "GET"))

    async def get_audio_volume(self) -> AudioVolume:
        """Get the current audio volume."""
        return AudioVolume(await self._request("/audio/volume", "GET"))

    async def set_audio_volume(self, volume) -> AudioVolume:
        """Set the audio volume."""
        body = {"value": volume}
        return AudioVolume(await self._request("/audio/volume", "PUT", body))

    async def get_now_playing(self) -> ContentNowPlaying:
        """Get the current playing content."""
        return ContentNowPlaying(await self._request("/content/nowPlaying", "GET"))

    async def get_bluetooth_status(self):
        """Get the Bluetooth status."""
        return await self._request("/bluetooth/source/status", "GET")

    async def get_power_state(self) -> SystemPowerControl:
        """Get the power state of the device."""
        return await self._request("/system/power/control", "GET")

    async def set_power_state(self, state: bool) -> None:
        """Set the power state of the device."""
        body = {"power": "ON" if state else "OFF"}
        await self._request("/system/power/control", "POST", body)

    async def _control_transport(self, control: str) -> ContentNowPlaying:
        """Control the transport."""
        body = {"state": control}
        return ContentNowPlaying(await self._request("/content/transportControl", "PUT", body))

    async def pause(self) -> ContentNowPlaying:
        """Pause the current content."""
        return await self._control_transport("PAUSE")

    async def play(self) -> ContentNowPlaying:
        """Play the current content."""
        return await self._control_transport("PLAY")

    async def skip_next(self) -> ContentNowPlaying:
        """Skip to the next content."""
        return await self._control_transport("SKIPNEXT")

    async def skip_previous(self) -> ContentNowPlaying:
        """Skip to the previous content."""
        return await self._control_transport("SKIPPREVIOUS")

    def get_device_id(self) -> str | None:
        """Get the device ID."""
        return self._device_id

    async def subscribe(self, resources: list[str] = DEFAULT_SUBSCRIBE_RESOURCES):
        """Subscribe to resources."""
        body = {"notifications": [{"resource": resource, "version": 1} for resource in resources ]}
        self._subscribed_resources = resources
        return await self._request("/subscription", "PUT", body, version=2)

    async def switch_tv_source(self) -> ContentNowPlaying:
        """Switch to TV source."""
        return await self.set_source("PRODUCT", "TV")

    async def set_source(self, source, sourceAccount) -> ContentNowPlaying:
        """Set the source."""
        body = {
            "source": source,
            "sourceAccount": sourceAccount
        }
        return ContentNowPlaying(await self._request("/content/playbackRequest", "POST", body))

    async def get_sources(self):
        """Get the sources."""
        return Sources(await self._request("/system/sources", "GET"))

    async def get_audio_setting(self, option) -> Audio:
        """Get the audio setting."""
        # TODO: load from capabilities
        if option not in ["bass", "treble", "center", "subwooferGain", "height", "avSync"]:
            raise Exception(f"Invalid audio setting: {option}")
        return Audio(await self._request("/audio/" + option, "GET"))

    async def set_audio_setting(self, option, value) -> Audio:
        """Get the audio setting."""
        # TODO: load from capabilities
        if option not in ["bass", "treble", "center", "subwooferGain", "height", "avSync"]:
            raise Exception(f"Invalid audio setting: {option}")

        return Audio(await self._request("/audio/" + option, "POST", {
            "value": int(value)
        }))

    async def get_accessories(self) -> Accessories:
        """Get the accessories."""
        return Accessories(await self._request("/accessories", "GET"))
    
    async def put_accessories(self, subs_enabled = None, rears_enabled = None) -> bool:
        if subs_enabled is None and rears_enabled is None:
            accessories = await self.get_accessories()
            if subs_enabled is None:
                subs_enabled = accessories.enabled.subs
            if rears_enabled is None:
                rears_enabled = accessories.enabled.rears
        
        body = {
            "enabled": {
                "rears": rears_enabled,
                "subs": subs_enabled
            }
        }
        return await self._request("/accessories", "PUT", body)

    async def get_battery_status(self) -> Battery:
        """Get the battery status."""
        return Battery(await self._request("/system/battery", "GET"))
        

# EXAMPLE USAGE

async def main(control_token, device_id, host):
    bose = BoseSpeaker(
        control_token=control_token,
        device_id=device_id,
        host=host
    )

    # Attach receiver for unsolicited messages
    bose.attach_receiver(lambda data: print(f"Received unsolicited message: {json.dumps(data, indent=4)}"))

    # Connect to the speaker
    await bose.connect()

    # Get system info
    response = await bose.get_system_info()
    print(response)

    # Get audio volume
    response = await bose.get_audio_volume()
    print(response)

    # Set get currently playing content
    response = await bose.get_now_playing()
    print(response)

    # Safely disconnect from the speaker
    await bose.disconnect()

if __name__ == "__main__":
    if len(sys.argv) != 4:
        print("Usage: python {sys.argv[0]} <control_token> <device_id> <host>")
        sys.exit(1)

    control_token = sys.argv[1]
    device_id = sys.argv[2]
    host = sys.argv[3]

    asyncio.run(main(control_token, device_id, host))


from typing import TypedDict, List, Optional, Dict


# SystemInfo
class SystemInfo(TypedDict):
    countryCode: str
    defaultName: str
    limitedFeatures: bool
    name: str
    productColor: int
    productId: int
    productName: str
    productType: str
    regionCode: str
    serialNumber: str
    softwareVersion: str
    variantId: int


# AudioVolume
class VolumeProperties(TypedDict):
    maxLimit: int
    maxLimitOverride: bool
    minLimit: int
    startupVolume: int
    startupVolumeOverride: bool


class AudioVolume(TypedDict):
    defaultOn: int
    max: int
    min: int
    muted: bool
    properties: VolumeProperties
    value: int


# ContentNowPlaying
class ContentItem(TypedDict, total=False):
    isLocal: bool
    presetable: bool
    source: str
    sourceAccount: str
    containerArt: str


class Capabilities(TypedDict, total=False):
    favoriteSupported: bool
    ratingsSupported: bool
    repeatSupported: bool
    resumeSupported: bool
    seekRelativeBackwardSupported: bool
    seekRelativeForwardSupported: bool
    shuffleSupported: bool
    skipNextSupported: bool
    skipPreviousSupported: bool


class Container(TypedDict, total=False):
    contentItem: Optional[ContentItem]
    capabilities: Optional[Capabilities]


class Source(TypedDict, total=False):
    sourceDisplayName: str
    sourceID: str


class Metadata(TypedDict, total=False):
    album: str
    artist: str
    duration: int
    trackName: str


class State(TypedDict, total=False):
    canFavorite: bool
    canPause: bool
    canRate: bool
    canRepeat: bool
    canSeek: bool
    canShuffle: bool
    canSkipNext: bool
    canSkipPrevious: bool
    canStop: bool
    quality: str
    repeat: str
    shuffle: str
    status: str
    timeIntoTrack: int
    timestamp: str


class Track(TypedDict, total=False):
    contentItem: Optional[ContentItem]
    favorite: str
    rating: str


class ContentNowPlaying(TypedDict, total=False):
    collectData: bool
    container: Optional[Container]
    source: Optional[Source]
    initiatorID: str
    metadata: Optional[Metadata]
    state: Optional[State]
    track: Optional[Track]


# System Power Control
class SystemPowerControl(TypedDict):
    power: str


# Sources
class SourceData(TypedDict, total=False):
    accountId: str
    displayName: str
    local: bool
    multiroom: bool
    sourceAccountName: str
    sourceName: str
    status: str
    visible: bool


class SourceProperties(TypedDict, total=False):
    supportedActivationKeys: List[str]
    supportedDeviceTypes: List[str]
    supportedFriendlyNames: List[str]
    supportedInputRoutes: List[str]


class Sources(TypedDict):
    properties: SourceProperties
    sources: List[SourceData]


# Audio
class AudioProperties(TypedDict, total=False):
    max: int
    min: int
    step: int
    supportedPersistence: bool


class Audio(TypedDict, total=False):
    persistence: bool
    properties: Optional[AudioProperties]
    value: int


# Accessories
class AccessoryData(TypedDict, total=False):
    available: bool
    configurationStatus: str
    serialnum: str
    type: str
    version: str
    wireless: bool


class Accessories(TypedDict, total=False):
    controllable: Optional[dict]
    enabled: Optional[dict]
    pairing: bool
    rears: Optional[List[AccessoryData]]
    subs: Optional[List[AccessoryData]]


# Battery
class Battery(TypedDict, total=False):
    chargeStatus: str
    chargerConnected: str
    minutesToEmpty: int
    minutesToFull: int
    percent: int
    sufficientChargerConnected: bool
    temperatureState: str


"""Bose cloud api responses:"""

# V4V Input
class V4VInput(TypedDict):
    input: str
    nameID: str


class Attributes(TypedDict, total=False):
    v4vInputs: List[V4VInput]


# Preset Metadata & Payload
class PresetMetadata(TypedDict):
    accountID: str
    image: str
    name: str
    subType: str


class ContentItem(TypedDict):
    containerArt: str
    location: str
    name: str
    presetable: bool
    source: str
    sourceAccount: str
    type: str


class PresetPayload(TypedDict):
    contentItem: ContentItem


class PresetAction(TypedDict):
    actionType: str
    metadata: PresetMetadata
    payload: PresetPayload


class Preset(TypedDict):
    actions: List[PresetAction]


# Service Accounts & Tokens
class ServiceAccountTokens(TypedDict, total=False):
    refresh_token: str
    refreshToken: str
    tokenType: str


class ServiceAccountAttributes(TypedDict, total=False):
    alexaEnv: str
    region: str
    WuWModel: str
    WuWord: str
    email: str
    language: str
    isDefaultAccount: bool


class ServiceAccount(TypedDict, total=False):
    accountID: str
    accountType: str
    bosePersonID: str
    createdOn: str
    provider: str
    providerAccountID: str
    tokens: ServiceAccountTokens
    updatedOn: str
    attributes: Optional[ServiceAccountAttributes]
    disabled: Optional[bool]
    name: Optional[str]
    productID: Optional[str]


# Users
class UserRole(TypedDict):
    role: str
    trustLevel: str


# Settings
class Settings(TypedDict):
    language: str
    name: str
    sharingMode: str
    timeFormat: str
    timeZone: str


# Main Data Structure
class BoseApiProduct(TypedDict):
    attributes: Attributes
    createdOn: str
    groups: List[str]
    persons: Dict[str, str]
    presets: Dict[str, Preset]
    productColor: int
    productID: str
    productType: str
    serviceAccounts: List[ServiceAccount]
    settings: Settings
    updatedOn: str
    users: Dict[str, UserRole]


import requests
import time
import json
import logging
import sys
import jwt
from .GSSDK import GSRequest, SigUtils
from .BoseResponse import BoseApiProduct

"""
ALL API KEYS ARE PUBLICLY AVAILABLE ON THE BOSE WEBSITE
THESE ARE NOT SENSITIVE INFORMATION AND THEREFORE CAN BE SHARED IN THIS SCRIPT

This script allows you to obtain a control token from the BOSE online api in order to control your BOSE speaker locally.
The obtained token is a JWT with a limited lifetime and needs to be refreshed from time to time. This script currently does not support refreshing the token. (Feel free to add it!)
So you need to refetch the token from time to time. Please cache the token, to avoid many api calls. I dont know if there is a rate limit or something like that.

This api was not documented by BOSE and is not officially supported, but was revered engineered by analyzing the BOSE app's api calls.
So it may stop working at any time.

IMPORTANT:
Just be respectful and dont spam the api with requests.
Otherwise BOSE may block this project all together.
"""

class BoseAuth:
    GIGYA_API_KEY = '3_7PoVX7ELjlWyppFZFGia1Wf1rNGZv_mqVgtqVmYl3Js-hQxZiFIU8uHxd8G6PyNz'
    GIGYA_UA = 'Bose/32768 MySSID/1568.300.101 Darwin/24.2.0'

    BOSE_API_KEY = '67616C617061676F732D70726F642D6D61647269642D696F73'

    def __init__(self):
        self._control_token = None

    def _get_ids(self):
        """
        Start a session and get the GMID and UCID
        """
        logging.debug("Getting GMID and UCID")

        url = "https://socialize.us1.gigya.com/socialize.getSDKConfig"
        data = {
            "apikey": self.GIGYA_API_KEY,
            "format": "json",
            "httpStatusCodes": False,
            "include": "permissions,ids,appIds",
            "sdk": "ios_swift_1.0.8",
            "targetEnv": "mobile"
        }
        try:
            response = requests.post(url, data=data).json()
        except Exception as e:
            logging.error(f"Error getting GMID and UCID: {e}")
            return None
        
        logging.debug(f"_get_ids: {json.dumps(response, indent=4)}")
        return {
            "gmid": response.get("ids", {}).get("gmid"),
            "ucid": response.get("ids", {}).get("ucid"),
        }

    def _login(self, email, password, gmid, ucid):
        """
        Login to Gigya
        """

        logging.debug(f"Logging in with {email}, gmid {gmid}, ucid {ucid}")
        url = 'https://accounts.us1.gigya.com/accounts.login'
        headers = {
            'Host': 'accounts.us1.gigya.com',
            'Connection': 'keep-alive',
            'Accept': '*/*',
            'User-Agent': self.GIGYA_UA,
            'Accept-Language': 'de-DE,de;q=0.9',
            'Content-Type': 'application/x-www-form-urlencoded',
        }

        data = {
            'apikey': self.GIGYA_API_KEY,
            'format': 'json',
            'gmid': gmid,
            'httpStatusCodes': 'false',
            'include': 'profile,data,emails,subscriptions,preferences,',
            'includeUserInfo': 'true',
            'lang': 'de',
            'loginID': email,
            'loginMode': 'standard',
            'password': password,
            'sdk': 'ios_swift_1.0.8',
            'sessionExpiration': '0',
            'source': 'showScreenSet',
            'targetEnv': 'mobile',
            'ucid': ucid,
        }
        response = requests.post(url, headers=headers, data=data)

        
        if response.status_code == 200:
            json_response = response.json()
            logging.debug("WARNING! CONFIDENTIAL INFORMATION! REMOVE AT LEAST THE session_secret AND UIDSignature FROM THE LOGS!")
            logging.debug(f"_login: {json.dumps(json_response, indent=4)}")
            logging.debug("END OF CONFIDENTIAL INFORMATION!")
            return {
                "session_token": json_response.get("sessionInfo", {}).get("sessionToken"),
                "session_secret": json_response.get("sessionInfo", {}).get("sessionSecret"),
                "uid": json_response.get("userInfo", {}).get("UID"),
                "signatureTimestamp": json_response.get("userInfo", {}).get("signatureTimestamp"),
                "UIDSignature": json_response.get("userInfo", {}).get("UIDSignature"),
            }
        else:
            raise ValueError(f"Login failed: {response.text}")

    def _get_jwt(self, user, gmid, ucid):
        """
        Get the authentication token from Gigya
        """

        url = "https://accounts.us1.gigya.com/accounts.getJWT"
        headers = {
            'Host': 'accounts.us1.gigya.com',
            'Connection': 'keep-alive',
            'Accept': '*/*',
            'User-Agent': self.GIGYA_UA,
            'Accept-Language': 'de-DE,de;q=0.9',
            'Content-Type': 'application/x-www-form-urlencoded',
        }

        timestamp = str(int(time.time()))
        params = {
            "apikey": self.GIGYA_API_KEY,
            "format": "json",
            "gmid": gmid,
            "httpStatusCodes": "false",
            "nonce": f"{timestamp}_1637928129",
            "oauth_token": user['session_token'],
            "sdk": "ios_swift_1.0.8",
            "targetEnv": "mobile",
            "timestamp": timestamp,
            "ucid": ucid,
        }

        request = GSRequest()
        base_string = request.calcOAuth1BaseString("POST", url, True, params)
        sig = SigUtils.calcSignature(base_string, user['session_secret'])
        params['sig'] = sig

        try:
            logging.debug("WAARNING! CONFIDENTIAL INFORMATION!")
            response = requests.post(url, headers=headers, data=params).json()
            logging.debug(f"_get_jwt: {json.dumps(response, indent=4)}")
            logging.debug("END OF CONFIDENTIAL INFORMATION!")
        except Exception as e:
            logging.error(f"Error getting JWT: {e}")
            return None
        return response.get("id_token")

    def _fetch_keys(self, gigya_jwt, signature_timestamp, uid, uid_signature):
        """
        Fetch the local control token
        """
        url = "https://id.api.bose.io/id-jwt-core/token"
        headers = {
            "X-ApiKey": self.BOSE_API_KEY,
            "X-Software-Version": "10.6.6-32768",
            "X-Api-Version": "1",
            "User-Agent": "MadridApp/10.6.6 (com.bose.bosemusic; build:32768; iOS 18.3.0) Alamofire/5.6.2"
        }
        data = {
            "id_token": gigya_jwt,
            "scope": "openid",
            "grant_type": "id_token",
            "signature_timestamp": signature_timestamp,
            "uid_signature": uid_signature,
            "uid": uid,
            "client_id": self.BOSE_API_KEY
        }

        try:
            response = requests.post(url, headers=headers, json=data).json()
            logging.debug("WARNING! CONFIDENTIAL INFORMATION!")
            logging.debug(f"_fetch_keys: {json.dumps(response, indent=4)}")
            logging.debug("END OF CONFIDENTIAL INFORMATION!")
        except Exception as e:
            logging.error(f"Error fetching keys: {e}")
            return None
        return response

    def is_token_valid(self, token):
        """
        Check if the token is still valid
        """
        try:
            decoded = jwt.decode(token, options={"verify_signature": False})
            exp = decoded.get("exp", 0)
            return exp > int(time.time())
        except Exception:
            return False

    # TODO: Implement refresh token
    def getControlToken(self, email, password, forceNew=False):
        """
        Get the control token to access the local speaker API
        """
        if not forceNew and self._control_token is not None:
            access_token = self._control_token.get("access_token")
            if access_token and self._is_token_valid(access_token):
                return {
                    "access_token": access_token,
                    "refresh_token": self._control_token.get("refresh_token"),
                }

        ids = self._get_ids()
        gmid, ucid = ids["gmid"], ids["ucid"]

        user = self._login(email, password, gmid, ucid)
        gigya_jwt = self._get_jwt(user, gmid, ucid)

        self._control_token = self._fetch_keys(
            gigya_jwt, user['signatureTimestamp'], user['uid'], user['UIDSignature']
        )
        return {
            "access_token": self._control_token.get("access_token"),
            "refresh_token": self._control_token.get("refresh_token"),
        }
        
    def fetchProductInformation(self, gwid) -> BoseApiProduct:
        url = "https://users.api.bose.io/passport-core/products/{}".format(gwid)
        headers = {
            "X-ApiKey": self.BOSE_API_KEY,
            "X-Software-Version": "10.6.6-32768",
            "X-Api-Version": "1",
            "User-Agent": "MadridApp/10.6.6 (com.bose.bosemusic; build:32768; iOS 18.3.0) Alamofire/5.6.2",
            "X-User-Token": self._control_token.get("access_token")
        }

        try:
            response = requests.get(url, headers=headers).json()
            logging.debug(f"product info: {json.dumps(response, indent=4)}")
        except Exception as e:
            logging.error(f"Error fetching keys: {e}")
            return None
        return BoseApiProduct(**response)

# EXAMPLE USAGE

if __name__ == "__main__":
    if len(sys.argv) != 3:
        print("Usage: python script.py <email> <password>")
        sys.exit(1)

    email = sys.argv[1]
    password = sys.argv[2]

    bose_auth = BoseAuth()
    control_token = bose_auth.getControlToken(email, password)
    print(json.dumps(control_token, indent=4))